@@ -5,6 +5,7 @@ from .CrossTabulate import cross_tabulate
 from .DocumentTermWeightFunctions import apply_term_weight_functions
 import pandas
 import scipy
+import warnings
 
 
 # ======================================================================
@@ -94,6 +95,16 @@ class SparseMatrixRecommender:
         """Take the pipeline value."""
         return self._value
 
+    def sub_matrix(self, tag_type):
+        """Take sub-matrix corresponding to tag_type."""
+        if not isinstance(tag_type, str):
+            raise ValueError("The first argument, tag_type, is expected to be a string.")
+
+        if not tag_type in self.take_matrices():
+            raise ValueError("The tag type " + str(tag_type) + " is not known in the recommender.")
+
+        return self.take_matrices()[tag_type]
+
     # ------------------------------------------------------------------
     # Echoers
     # ------------------------------------------------------------------
@@ -394,7 +405,7 @@ class SparseMatrixRecommender:
     # ------------------------------------------------------------------
     # Recommend by profile
     # ------------------------------------------------------------------
-    def recommend_by_profile(self, profile, nrecs=10, vector_result: bool = False):
+    def recommend_by_profile(self, profile, nrecs=10, normalize=True, vector_result: bool = False):
         """Recommend by profile.
 
         :type profile: str|list|dict
@@ -403,6 +414,9 @@ class SparseMatrixRecommender:
         :type nrecs: int|None
         :param nrecs: A positive integer or None. If it is None, then all items with non-zero scores are returned.
 
+        :type normalize: bool
+        :param normalize: Should the result be normalized or not.
+
         :type vector_result: bool
         :param vector_result: Should the result be a (SSparseMatrix) vector or a dictionary.
 
@@ -422,7 +436,27 @@ class SparseMatrixRecommender:
         # Compute the recommendations
         recs = self.take_M().dot(vec)
 
-        if not vector_result:
+        # Normalize
+        recs_max = max(map(abs, recs.row_sums()))
+        if normalize and recs_max > 0:
+            recs = recs.multiply(1 / recs_max)
+
+        if vector_result:
+            # Vector result
+
+            # Change the matrix to have nrecs columns
+            # according to scores
+            if isinstance(nrecs, int) and nrecs < recs.rows_count():
+                recs2 = recs.row_sums_dict()
+                recs2 = _reverse_sort_dict(recs2)
+
+                recs2 = dict(list(recs2.items())[0:nrecs])
+                rowNames = recs.row_names()
+                recs = recs[list(recs2.keys()), :]
+                recs = recs.impose_row_names(rowNames)
+
+        else:
+            # Dictionary result
             # Take non-zero score recommendations
             recs = {key: value for (key, value) in recs.row_sums_dict().items() if value > 0}
 
@@ -443,7 +477,7 @@ class SparseMatrixRecommender:
     # ------------------------------------------------------------------
     # Recommend by history
     # ------------------------------------------------------------------
-    def recommend(self, history, nrecs=10, remove_history=True):
+    def recommend(self, history, nrecs=10, normalize=True, remove_history=True):
         """Recommend by history.
 
         :type history: str|list|dict
@@ -452,6 +486,9 @@ class SparseMatrixRecommender:
         :type nrecs: int|None
         :param nrecs: A positive integer or None. If it is None, then all items with non-zero scores are returned.
 
+        :type normalize: bool
+        :param normalize: Should the result be normalized or not.
+
         :type remove_history: bool
         :param remove_history: Should the the history be removed from the result recommendations or not?
 
@@ -488,6 +525,12 @@ class SparseMatrixRecommender:
         # Give top-n recs
         if isinstance(nrecs, int) and nrecs < len(recs):
             recs = dict(list(recs.items())[0:nrecs])
+
+            # Normalize
+            recs_max = max(map(abs, recs.values()))
+            if normalize and recs_max > 0:
+                recs = {k: v / recs_max for (k, v) in recs.items()}
+
         elif not (isinstance(None, type(None)) or isinstance(nrecs, int) and nrecs > 0):
             raise TypeError("The second argument, nrecs, is expected to be a positive integer or None.")
 
@@ -585,7 +628,7 @@ class SparseMatrixRecommender:
                             drop_zero_scored_labels=True,
                             max_number_of_labels=None,
                             normalize: bool = True):
-        """Clasify by profile vector.
+        """Classify by profile vector.
 
         :type tag_type: str
         :param tag_type: Tag type to classify to.
@@ -618,6 +661,7 @@ class SparseMatrixRecommender:
         recs = self.recommend_by_profile(profile=profile,
                                          nrecs=n_top_nearest_neighbors,
                                          vector_result=True).take_value()
+
         # "Nothing" result
         if recs.column_sums()[0] == 0:
             self.set_value({None: 1})
@@ -661,6 +705,102 @@ class SparseMatrixRecommender:
 
         return self
 
+    # ------------------------------------------------------------------
+    # To metadata recommender
+    # ------------------------------------------------------------------
+    def to_metadata_recommender(self, tag_type_to, tag_types=None, tag_type_matrix=None, normalizer_func=None):
+        """Convert to a metadata recommender
+
+         Converts the recommender object into a recommender for a
+         specified tag type using replacements in the long form representation of the recommender matrix.
+
+         The following steps are taken:
+
+           - If `tag_type_matrix` is None then tag_type_matrix = self.sub_matrix(tag_type_to)
+
+           - Normalize the columns of `tag_type_matrix` using `normalizer_func`
+
+           - Each recommender sub-matrix `m` is multiplied by `tag_type_matrix`, i.e. `tag_type_matrix.transpose().dot(m)`
+
+           - A new recommender is created with items that are the tags of `tag_type_to`
+
+         :type tag_type_to: Str
+         :param tag_type_to: Tag type to make a recommender for.
+
+         :type tag_types: List
+         :param tag_types: A vector tag types (strings) to make the data frame with. If None then all tag types are used.
+
+         :type tag_type_matrix: SSparseMatrix|None
+         :param tag_type_matrix: A sparse matrix of item IDs vs tags of tag_type_to.
+         If None then self.sub_matrix(tag_type_to) is used.
+
+         :type normalizer_func:
+         :param normalizer_func: An LSI normalizer function. One of None, "None", "Cosine", "Sum", or "Max".
+         If None then it is same as "None". See self.apply_term_weight_functions.
+
+         :rtype: SparseMatrixRecommender
+         :return A sparse matrix recommender
+         """
+
+        # Process tag_type_to
+        if not isinstance(tag_type_to, str):
+            raise ValueError("The argument tag_type_to is expected to be a string.")
+
+        # Verify tag_type_to
+        if tag_type_to not in self.take_matrices():
+            raise ValueError("The value of the first argument, 'tag_type_to' is not a known tag type.")
+
+        my_tag_type_to = tag_type_to
+
+        # Process tag_types
+        allTagTypes = set(self.take_matrices().keys())
+        my_tag_types = tag_types
+        if tag_types is None:
+            my_tag_types = allTagTypes.difference({my_tag_type_to})
+
+        if my_tag_type_to in set(my_tag_types):
+            warnings.warn("Removing the value of tag_type_to from value of tag_types.")
+            my_tag_types = my_tag_types.difference({my_tag_type_to})
+
+        if len(allTagTypes.intersection(my_tag_types)) == 0:
+            raise ValueError("The argument tag_types has no known tag type of the recommender object.")
+
+        # Process tag_type_matrix
+        my_tag_type_matrix = tag_type_matrix
+        if not (tag_type_matrix is None or isinstance(tag_type_matrix, SSparseMatrix)):
+            raise ValueError("The argument tag_type_matrix is expected to be None or a sparse matrix.")
+
+        if tag_type_matrix is None:
+            my_tag_type_matrix = self.sub_matrix(tag_type_to)
+
+        if not my_tag_type_matrix.row_names() == self.take_M().row_names():
+            raise ValueError(
+                "The argument tag_type_matrix has row names that are different than the row names of recommender matrix.")
+
+        # Process normalizer_func
+        my_normalizer_func = normalizer_func
+        if not (normalizer_func is None or isinstance(normalizer_func, str)):
+            raise ValueError("The argument normalizerFunc is expected to be None or a string.")
+
+        # Obtain tag_type_matrix
+        my_tag_type_matrix = my_tag_type_matrix.transpose()
+
+        # Normalize the columns of tagTypeMatrix
+        if not (normalizer_func is None or normalizer_func == "None"):
+            my_tag_type_matrix = apply_term_weight_functions(doc_term_matrix=my_tag_type_matrix,
+                                                             global_weight_func="None",
+                                                             local_weight_func="None",
+                                                             normalizer_func=my_normalizer_func)
+
+        # Get the recommender contingency matrices
+        smats = {k: m for (k, m) in self.take_matrices().items() if k in my_tag_types}
+
+        # Multiply sub-matrices
+        smats = {k: my_tag_type_matrix.dot(m) for (k, m) in smats.items()}
+
+        # Create recommender
+        return SparseMatrixRecommender().create_from_matrices(matrices=smats)
+
     # ------------------------------------------------------------------
     # To dictionary form
     # ------------------------------------------------------------------
